#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#define MAXDELAY	5
#define MAXREPEAT	5
#define BUFSIZE		128
#define SEQFILE		"seqno"
#define MAXTRIES	5
#define WAITTIME	1
#define P(sem)		if (!my_lock(sem)) \
			    perror("my_lock"), exit(6);
#define V(sem)		if (!my_unlock(sem)) \
			    perror("my_unlock"), exit(7);

/* Zugriff ueber die Semaphore sem (Filedeskriptor) synchronisieren
 * Rueckgabewert: != 0 bei Erfolg
 */
int my_lock(int sem) {
    int ret, tries = 0;
    /* zuerst auf den Anfang der Datei positionieren und dann mittels
     * lockf() bis zum Dateiende alles sperren
     */
    while ((ret = lseek(sem, 0, SEEK_SET)) == 0		/* an den Anfang gehen ... */
           && (ret = lockf(sem, F_TLOCK, 0)) < 0	/* ... und bis zum Ende sperren */
	   && (errno == EACCES || errno == EAGAIN)) {
	if (++tries > MAXTRIES)	/* ... das ganze nur MAXTRIES Mal versuchen */
	    return 0;
	sleep(WAITTIME);	/* ... und dazwischen immer ein kleines Schlaefchen */
    }
    return ret == 0;
}

/* Semaphore wieder freigeben (d.h. Datei loeschen)
 * Rueckgabewert: != 0 bei Erfolg
 */
int my_unlock(int sem) {
    return lseek(sem, 0, SEEK_SET) == 0		/* an den Anfang gehen ... */
	   && lockf(sem, F_ULOCK, 0) == 0;	/* ... und bis zum Ende freigeben */
}

/* Zaehler lesen, delay Sekunden schlafen und dann erhoehen */
void incr(int delay) {
    int fd, seqno, n;
    char buf[BUFSIZE];
    
    if ((fd = open(SEQFILE, O_RDWR)) < 0)
	perror("open"), exit(2);

    P(fd)	/* kritischen Bereich beginnen; Lock anfordern */
    
    puts("BEGIN kritischer Bereich");

	/* Dateiinhalt in den Puffer einlesen */
    if ((n = read(fd, buf, BUFSIZE)) < 0)
	perror("read"), exit(3);
	/* ... und mit Null-Byte terminieren */
    buf[n] = '\0';
    
	/* Integer parsen */
    if (sscanf(buf, "%d", &seqno) != 1)
	fprintf(stderr, "sscanf error\n"), exit(4);

    printf("gelesen: %d\n", seqno);

    seqno++;	/* Zaehler erhoehen */

    sleep(delay);	/* delay Sekunden schlafen */
    
    sprintf(buf, "%03d\n", seqno);	/* neuen Zaehler als String */
    n = strlen(buf);
    
    lseek(fd, 0, SEEK_SET);		/* an den Dateianfang gehen */
    
    if (write(fd, buf, n) != n)		/* ... und den Zaehler schreiben */
	perror("write"), exit(5);
    
    puts("END kritischer Bereich");
    
    V(fd)	/* kritischen Bereich verlassen; Lock wieder freigeben */

    close(fd);
}

int main(int argc, char **argv) {
    int i, delay;

    if (argc != 2) {
	fprintf(stderr, "Usage: %s <seed>\n", argv[0]);
	exit(1);
    }

    /* Pseudo-Zufallszahlengenerator initialisieren */
    srandom(atoi(argv[1]));

    /* zufaellig zwischen 0 und MAXDELAY Sekunden schlafen */
    sleep(delay = random() % (MAXDELAY+1));
    
    for (i = 1; i <= MAXREPEAT; i++) {
    	/* gemeinsamen Zaehler erhoehen */
	incr(delay);
	/* zufaellig zwischen 0 und MAXDELAY Sekunden schlafen */
	sleep(delay = random() % (MAXDELAY+1));
    }

    return 0;
}
